from typing import Any

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import optax
from jaxtyping import Array, Bool, Float, Int, Scalar


class _OptimizerState(eqx.Module, strict=True):
    step: Int[Array, ""]
    old_loss: Scalar
    curr_loss: Scalar
    opt_state: tuple[Any, Any]


class _AtomicModel(eqx.Module):
    positions: Float[Array, "n_atoms 3"]
    amplitudes: Float[Array, "n_atoms n_gaussians"]
    variances: Float[Array, "n_atoms n_gaussians"]


class _Volume(eqx.Module):
    voxel_grid: Float[Array, "z y x"]
    voxel_size: float


class Solution(eqx.Module):
    rotation_matrix: Float[Array, "3 3"]
    offset: Float[Array, " 3"]
    correlation: Float
    n_steps: Int


class ModelToVolumeAligner(eqx.Module):
    volume: _Volume
    optimizers: tuple[Any, Any]
    rtol: float
    atol: float
    max_steps: int

    def __init__(
        self,
        real_voxel_grid: Float[Array, "z y x"],
        voxel_size: float,
        rtol: float = 1e-5,
        atol: float = 1e-5,
        max_steps: int = 100,
        optimizers: tuple[Any, Any] | None = None,
    ):
        assert voxel_size > 0, "voxel_size must be positive."
        assert max_steps > 0, "max_steps must be positive."
        if optimizers is None:
            optimizers = (
                optax.adabelief(learning_rate=1e-1, nesterov=True),
                optax.adabelief(learning_rate=1e-1, nesterov=True),
            )
        self.optimizers = optimizers
        self.rtol = rtol
        self.atol = atol
        self.max_steps = int(max_steps)
        self.volume = _Volume(voxel_grid=real_voxel_grid, voxel_size=voxel_size)

    @eqx.filter_jit
    def _init(
        self,
        quat_init: Float[Array, "4"],
        offset_init: Float[Array, "3"],
    ) -> _OptimizerState:
        opt_state = (
            self.optimizers[0].init(quat_init),
            self.optimizers[1].init(offset_init),
        )
        maxval = jnp.array(jnp.finfo(quat_init.dtype).max, quat_init.dtype)
        return _OptimizerState(
            step=jnp.array(0), old_loss=maxval, curr_loss=maxval, opt_state=opt_state
        )

    @eqx.filter_jit
    def _step(
        self,
        val: tuple[Float[Array, "4"], Float[Array, "3"], _OptimizerState],
        args: tuple[_AtomicModel, _Volume],
    ) -> tuple[Float[Array, "4"], Float[Array, "3"], _OptimizerState]:
        quat, offset, state = val
        q_opt_state, d_opt_state = state.opt_state
        optim_q, optim_d = self.optimizers

        loss, grad_q = _loss_and_grad_quat_fn(quat, offset, args)
        updates_q, q_opt_state = optim_q.update(grad_q, q_opt_state)
        quat = optax.apply_updates(quat, updates_q)
        quat /= jnp.linalg.norm(quat, keepdims=True)

        loss, grad_d = _loss_and_grad_offset_fn(quat, offset, args)
        updates_d, d_opt_state = optim_d.update(grad_d, d_opt_state)
        offset = jnp.asarray(optax.apply_updates(offset, updates_d))

        step = state.step + 1
        new_state = _OptimizerState(
            step=step,
            old_loss=state.curr_loss,
            curr_loss=loss,
            opt_state=(q_opt_state, d_opt_state),
        )

        # jax.debug.print("Step={step}, Corr={loss}", step=step, loss=1 - loss)
        return quat, offset, new_state

    @eqx.filter_jit
    def _terminate_condition(self, state: _OptimizerState) -> Bool[Array, ""]:
        cond1 = jnp.abs(state.old_loss - state.curr_loss) > (
            self.atol + self.rtol * jnp.abs(state.curr_loss)
        )
        cond2 = state.step < self.max_steps
        return jax.lax.cond(state.step < 1, lambda x: True, lambda x: cond1 & cond2, None)

    @eqx.filter_jit
    def align(
        self,
        atomic_positions_in_angstroms: Float[Array, "n_atoms 3"],
        amplitudes: Float[Array, "n_atoms n_gaussians"],
        variances: Float[Array, "n_atoms n_gaussians"],
        quat_init: Float[Array, "4"] = jnp.array([1.0, 0.0, 0.0, 0.0]),
        offset_init: Float[Array, "3"] = jnp.array([0.0, 0.0, 0.0]),
    ) -> tuple[Float[Array, "n_atoms 3"], Solution]:
        atomic_model = _AtomicModel(
            positions=atomic_positions_in_angstroms,
            amplitudes=amplitudes,
            variances=variances,
        )

        args = (atomic_model, self.volume)
        state = self._init(quat_init, offset_init)

        init_val = (quat_init, offset_init, state)
        quat_opt, offset_opt, final_state = jax.lax.while_loop(
            lambda val: self._terminate_condition(val[2]),
            body_fun=lambda val: self._step(val, args),
            init_val=init_val,
        )

        offset_opt = jnp.asarray(offset_opt)

        rot_matrix = cxs.QuaternionPose(wxyz=quat_opt).rotation.as_matrix()

        solution = Solution(
            rotation_matrix=rot_matrix,
            offset=offset_opt,
            correlation=1 - final_state.curr_loss,
            n_steps=final_state.step,
        )

        aligned_positions = atomic_model.positions @ rot_matrix + offset_opt

        return aligned_positions, solution


def _atom_potential_to_volume(
    atom_positions: Float[Array, "n_atoms_atoms 3"],
    gaussian_amp: Float[Array, "n_atoms_atoms n_gaussians"],
    gaussian_var: Float[Array, "n_atoms_atoms n_gaussians"],
    shape: tuple[int, int, int],
    voxel_size: float,
) -> Float[Array, "z x y"]:
    atom_potential = cxs.GaussianMixtureVolume(atom_positions, gaussian_amp, gaussian_var)
    render_fn = cxs.GaussianMixtureRenderFn(shape, voxel_size)

    volume = render_fn(atom_potential)
    return volume


@eqx.filter_jit
def _loss_fn(
    quat: Float[Array, " 4"],
    offset: Float[Array, "3"],
    args: tuple[_AtomicModel, _Volume],
) -> Float[Array, ""]:
    atomic_model, volume = args

    rot_matrix = cxs.QuaternionPose(
        wxyz=quat,
    ).rotation.as_matrix()

    v_rot = _atom_potential_to_volume(
        atomic_model.positions @ rot_matrix + offset,
        atomic_model.amplitudes,
        atomic_model.variances,
        shape=volume.voxel_grid.shape,
        voxel_size=volume.voxel_size,
    )
    v_rot /= v_rot.sum(keepdims=True)

    return 1 - jnp.sum(v_rot * volume.voxel_grid) / jnp.linalg.norm(
        v_rot
    ) / jnp.linalg.norm(volume.voxel_grid)


@eqx.filter_jit
def _loss_and_grad_quat_fn(
    quat: Float[Array, " 4"],
    offset: Float[Array, "3"],
    args: tuple[_AtomicModel, _Volume],
) -> tuple[Float[Array, ""], Float[Array, " 4"]]:
    # print("compiling!")
    return jax.value_and_grad(_loss_fn, argnums=0)(quat, offset, args)


@eqx.filter_jit
def _loss_and_grad_offset_fn(
    quat: Float[Array, " 4"],
    offset: Float[Array, "3"],
    args: tuple[_AtomicModel, _Volume],
) -> tuple[Float[Array, ""], Float[Array, " 3"]]:
    # print("compiling!")
    return jax.value_and_grad(_loss_fn, argnums=1)(quat, offset, args)



from ._config_loader import load_config as load_config
from ._config_validators import (
    DatasetSimulatorConfig as DatasetSimulatorConfig,
    EnsOptMDConfig as EnsOptMDConfig,
    GMMFitConfig as GMMFitConfig,
)



import cryojax.simulator as cxs
from jaxtyping import Array, Float


def make_image_model_from_gmm(
    walker: Float[Array, "n_atoms 3"],
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
    variances: Float[Array, "n_atoms n_gaussians_per_atom"],
    image_config: cxs.BasicImageConfig,
    pose: cxs.AbstractPose,
    transfer_theory: cxs.ContrastTransferTheory,
) -> cxs.AbstractImageModel:
    volume = cxs.GaussianMixtureVolume(
        walker,
        amplitudes,
        variances,
    )
    return cxs.make_image_model(
        volume,
        image_config,
        pose,
        transfer_theory,
        normalizes_signal=True,
    )



from ._brownian_dynamics import (
    OverdampedLangevinSampler as OverdampedLangevinSampler,
    ParallelSteeredOverdampedLangevinSampler as ParallelSteeredOverdampedLangevinSampler,
    SteeredOverdampedLangevinSampler as SteeredOverdampedLangevinSampler,
)



from .data_generator_config import DatasetSimulatorConfig as DatasetSimulatorConfig
from .ensemble_opt_config import EnsOptMDConfig as EnsOptMDConfig
from .gmm_fit_config import GMMFitConfig as GMMFitConfig



import cryojax.simulator as cxs
import jax
import jax.numpy as jnp
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import CircularCosineMask

from cryojax_eo.simulator import DilatedMask, simulate_image_with_white_gaussian_noise


def test_dilated_mask_projection(sample_path_mrc_file):
    voxel_grid, voxel_size = read_array_from_mrc(
        sample_path_mrc_file, loads_grid_spacing=True
    )

    mask = jnp.where(jnp.abs(voxel_grid) > 0.001, 1.0, 0.0)

    image_config = cxs.BasicImageConfig(
        shape=(32, 32), pixel_size=voxel_size, voltage_in_kilovolts=300.0
    )
    dilated_mask = DilatedMask(voxel_grid, image_config=image_config)

    proj_mask = dilated_mask.project(cxs.EulerAnglePose())

    assert (
        jnp.abs(proj_mask - mask.sum(0)).min() < 1e-5
    ), "Projected mask does not match expected mask"

    return


def test_image_simulation(sample_path_to_pdb1, sample_path_to_pdb2):
    particle_parameters = {
        "pose": cxs.EulerAnglePose(),
        "image_config": cxs.BasicImageConfig(
            shape=(32, 32), pixel_size=0.8, voltage_in_kilovolts=300.0
        ),
        "transfer_theory": cxs.ContrastTransferTheory(
            cxs.AstigmaticCTF(defocus_in_angstroms=200, spherical_aberration_in_mm=1e-16),
            amplitude_contrast_ratio=0.1,
        ),
    }

    volumes = tuple(
        [
            cxs.load_tabulated_volume(
                filename,
                output_type=cxs.GaussianMixtureVolume,
                include_b_factors=True,
                selection_string="not element H",
            )
            for filename in [sample_path_to_pdb1, sample_path_to_pdb2]
        ]
    )

    mask = CircularCosineMask(
        particle_parameters["image_config"].coordinate_grid_in_pixels,
        radius=32 // 2,
        rolloff_width=1.0,
    )

    simulate_image_with_white_gaussian_noise(
        particle_parameters,
        constant_args=(volumes, mask, 1.0),
        per_particle_args=(jax.random.key(0), 0, 0.1),
    )

    return



import os
import shutil

import cryojax.simulator as cxs
import jax
import jax.numpy as jnp
import mdtraj
import pytest
from cryospax import RelionParticleDataset, RelionParticleParameterFile
from optax import constant_schedule

from cryojax_eo.dataset import create_dataloader
from cryojax_eo.ensemble_optimization import (
    EnsembleOptimizationPipeline,
    EnsembleSteeredMDSimulator,
    ImagesToEnsembleLikelihoodFn,
    IterativeEnsembleLikelihoodOptimizer,
    MargGaussianWhiteLogLikelihoodFn,
    SteeredMDSimulator,
)
from cryojax_eo.io import read_walkers_from_pdbs
from cryojax_eo.utils import ModelToVolumeAligner


@pytest.fixture
def sample_model_aligner(sample_path_to_pdb1):
    gmm_volume = cxs.load_tabulated_volume(
        sample_path_to_pdb1,
        output_type=cxs.GaussianMixtureVolume,
        selection_string="not element H",
        include_b_factors=True,
    )

    voxel_size = 0.2 * 128 / 32
    render_fn = cxs.GaussianMixtureRenderFn((32, 32, 32), voxel_size)
    real_voxel_grid = render_fn(gmm_volume)
    return ModelToVolumeAligner(real_voxel_grid, voxel_size)


def make_steered_md_simulator(path_to_pdb):
    model = mdtraj.load(path_to_pdb)
    atom_list = model.topology.select("not element H")

    return SteeredMDSimulator(
        path_to_pdb,
        n_steps=10,
        restrain_atom_list=atom_list,
        parameters_for_md={"platform": "CPU", "properties": {"Threads": "4"}},
        base_state_file_path=os.path.join(
            os.path.dirname(__file__), "outputs/md_states", "state_it"
        ),
    )


def test_ensemble_optimization_optimizer(
    sample_path_to_pdb1,
    sample_path_to_pdb2,
    sample_path_to_relion_project,
    sample_path_to_starfile,
    sample_model_aligner,
):
    prealigned_structure = mdtraj.load(sample_path_to_pdb1).center_coordinates()
    atom_list = prealigned_structure.topology.select("not element H")

    walkers, amplitudes, variances = read_walkers_from_pdbs(
        [sample_path_to_pdb1, sample_path_to_pdb2]
    )
    variances = variances[atom_list]
    amplitudes = amplitudes[atom_list]

    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            sample_path_to_starfile, options=dict(broadcasts_image_config=True)
        ),
        sample_path_to_relion_project,
    )

    dataloader = create_dataloader(
        relion_dataset,
        batch_size=2,
        shuffle=True,
        per_particle_args=None,
        jax_prng_key=jax.random.key(0),
    )

    img_to_walker_likelihood_fn = MargGaussianWhiteLogLikelihoodFn(
        amplitudes=amplitudes, variances=variances, image_sign=1.0, dilated_mask=None
    )
    ensemble_likelihood_fn = ImagesToEnsembleLikelihoodFn(img_to_walker_likelihood_fn)

    optimizer = IterativeEnsembleLikelihoodOptimizer(
        step_size=1.0,
        n_steps=2,
        n_batches_per_step=2,
        ensemble_likelihood_fn=ensemble_likelihood_fn,
        pose_search=None,
    )

    projector = EnsembleSteeredMDSimulator(
        [
            make_steered_md_simulator(sample_path_to_pdb1),
            make_steered_md_simulator(sample_path_to_pdb2),
        ]
    )

    pipeline = EnsembleOptimizationPipeline(
        projector,
        optimizer,
        n_steps=2,
        prealigned_structure=prealigned_structure,
        atom_indices_for_opt=atom_list,
        model_to_volume_aligner=sample_model_aligner,
        runs_postprocessing=True,
    )

    weights = jnp.array([0.5, 0.5])
    output_directory = os.path.join(
        os.path.dirname(__file__), "outputs/ensemble_optimization_test"
    )
    os.makedirs(output_directory, exist_ok=True)

    pipeline.run(
        jax.random.key(0),
        walkers,
        weights,
        dataloader,
        constant_schedule(1.0e3),
        output_directory=output_directory,
    )
    shutil.rmtree(os.path.join(os.path.dirname(__file__), "outputs/"))

    return



from ._likelihood_optimization import (
    AbstractEnsembleParameterOptimizer as AbstractEnsembleParameterOptimizer,
    GaussianWhiteLogLikelihoodFn as GaussianWhiteLogLikelihoodFn,
    ImagesToEnsembleLikelihoodFn as ImagesToEnsembleLikelihoodFn,
    IterativeEnsembleLikelihoodOptimizer as IterativeEnsembleLikelihoodOptimizer,
    MargGaussianWhiteLogLikelihoodFn as MargGaussianWhiteLogLikelihoodFn,
    ProjGradDescWeightOptimizer as ProjGradDescWeightOptimizer,
    compute_optimal_scale_and_offset as compute_optimal_scale_and_offset,
    make_image_model_from_gmm as make_image_model_from_gmm,
)
from ._pipelines import (
    AbstractEnsembleOptimizationPipeline as AbstractEnsembleOptimizationPipeline,
    EnsembleOptimizationPipeline as EnsembleOptimizationPipeline,
)
from ._pose_search import (
    HierarchicalSO3GridSearch as HierarchicalSO3GridSearch,
    compute_correlation_at_optimal_offset as compute_correlation_at_optimal_offset,
)
from ._prior_projection import (
    AbstractEnsemblePriorProjector as AbstractEnsemblePriorProjector,
    AbstractForceField as AbstractForceField,
    AbstractPriorProjector as AbstractPriorProjector,
    EnsembleSteeredMDSimulator as EnsembleSteeredMDSimulator,
    OverdampedLangevinSampler as OverdampedLangevinSampler,
    ParallelSteeredOverdampedLangevinSampler as ParallelSteeredOverdampedLangevinSampler,
    SteeredMDSimulator as SteeredMDSimulator,
    SteeredOverdampedLangevinSampler as SteeredOverdampedLangevinSampler,
    compute_harmonic_steering_force as compute_harmonic_steering_force,
)



from typing import Any

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from cryojax.jax_util import filter_bmap
from jaxtyping import Array, Float

from .single_likelihood import AbstractImageToWalkerLogLikelihoodFn


class AbstractImagesToEnsembleLikelihoodFn(eqx.Module):
    image_to_walker_likelihood_fn: eqx.AbstractVar[AbstractImageToWalkerLogLikelihoodFn]
    n_walkers_in_parallel: eqx.AbstractVar[int]
    n_images_in_parallel: eqx.AbstractVar[int]

    def compute_log_likelihood_matrix(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float[Array, "n_images n_walkers"]:
        raise NotImplementedError

    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        weights: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theory: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ):
        raise NotImplementedError


class ImagesToEnsembleLikelihoodFn(AbstractImagesToEnsembleLikelihoodFn):
    image_to_walker_likelihood_fn: AbstractImageToWalkerLogLikelihoodFn
    n_walkers_in_parallel: int = 1
    n_images_in_parallel: int = 1

    def compute_log_likelihood_matrix(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float[Array, "n_images n_walkers"]:
        vmap_over_images_fn = eqx.filter_vmap(
            lambda w, img, ic, p, tf, ppa: self.image_to_walker_likelihood_fn(
                w, img, ic, p, tf, ppa
            ),
            in_axes=(
                None,
                0,
                eqx.if_array(0),
                0,
                0,
                0,
            ),
        )

        vmap_over_walkers_fn = eqx.filter_vmap(
            lambda w, img, ic, p, tf, ppa: filter_bmap(
                f=lambda x: vmap_over_images_fn(w, x[0], x[1], x[2], x[3], x[4]),
                xs=(img, ic, p, tf, ppa),
                batch_size=self.n_images_in_parallel,
            ),
            in_axes=(0, None, None, 0, None, None),
        )
        return filter_bmap(
            lambda x: vmap_over_walkers_fn(
                x[0], images, image_config, x[1], transfer_theories, per_particle_args
            ),
            (walkers, poses_per_walker),
            batch_size=self.n_walkers_in_parallel,
        ).T

    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        weights: Float[Array, "n_walkers n_atoms n_gaussians_per_atom"],
        images: Float[Array, "n_images y x"],
        image_config: cxs.BasicImageConfig,
        poses_per_walker: cxs.AbstractPose,
        transfer_theories: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ):
        likelihood_matrix = self.compute_log_likelihood_matrix(
            walkers,
            images,
            image_config,
            poses_per_walker,
            transfer_theories,
            per_particle_args,
        )
        return jnp.mean(
            jax.scipy.special.logsumexp(a=likelihood_matrix, b=weights[None, :], axis=1)
        )



"""Creating commands installed with Cryo-MD for use from command line.

See the `[project.scripts]` entry in the `pyproject.toml` file for how this module
is used to create the commands during installation.

"""  # noqa: E501

import argparse
import os
from importlib import import_module

from .cryojax_ensemble_optimization_version import __version__


def _get_commands(cmd_dir: str, doc_str: str = "") -> None:
    """Start up a parser using the modules directory as subparsers.

    Arguments
    ---------
        cmd_dir: path to folder containing Cryo-MD command modules
        doc_str: documentation for this list of commands as a whole

    """  # noqa: E501
    parser = argparse.ArgumentParser(description=doc_str)
    parser.add_argument(
        "--version",
        action="version",
        version="cryojax-ensemble-optimization " + __version__,
    )

    subparsers = parser.add_subparsers(title="Choose a command")
    subparsers.required = True
    module_files = os.listdir(cmd_dir)
    dir_lbl = os.path.basename(cmd_dir)

    # look for Python modules that have the "add_args" method defined, which is what we
    # use to mark a module in these directories as added to the command namespace
    for module_file in module_files:
        if module_file != "__init__.py" and module_file[-3:] == ".py":
            module_name = ".".join(
                ["cryojax_ensemble_optimization", dir_lbl, module_file[:-3]]
            )
            module = import_module(module_name)

            if hasattr(module, "add_args"):
                parsed_doc = module.__doc__.split("\n") if module.__doc__ else list()
                descr_txt = parsed_doc[0] if parsed_doc else ""
                epilog_txt = "" if len(parsed_doc) <= 1 else "\n".join(parsed_doc[1:])

                # we add documentation text parsed from the module's docstring
                this_parser = subparsers.add_parser(
                    module_file[:-3],
                    description=descr_txt,
                    epilog=epilog_txt,
                    formatter_class=argparse.RawTextHelpFormatter,
                )
                module.add_args(this_parser)
                this_parser.set_defaults(func=module.main)

    args = parser.parse_args()
    args.func(args)


def main_commands():
    _get_commands(
        cmd_dir=os.path.join(os.path.dirname(__file__), "commands"),
        doc_str="Commands installed with cryojax-ensemble-opt package.\n\n",
    )



import cryojax.ndimage as cxim
import jax.numpy as jnp
from jaxtyping import Array, Float


def compute_correlation_at_optimal_offset(
    image_shifted: Float[Array, "H W"],
    image_ref: Float[Array, "H W"],
    coordinate_grid_in_angstroms: Float[Array, "H W 2"],
) -> tuple[Float[Array, ""], Float[Array, "2"]]:
    """
    Estimates the correlation between two images at the optimal shift
    using Plancherel's theorem and Fourier's convolution theorem.

    **Arguments:**
    - `image_shifted`:
        Image that has been shifted.
    - `image_ref`:
        Reference centered image.
    - `coordinate_grid_in_angstroms`:
        Coordinate grid that maps pixel indices to positions in angstroms.
    **Returns:**
    - `loss`:
        Negative correlation at the optimal shift.
    - `shift`:
        Shift that best aligns ``image_shifted`` to
        ``image_ref`` in the phase-correlation sense.
    """
    abs_cross_corr = jnp.abs(
        _compute_cross_correlation_image(
            image_shifted / jnp.linalg.norm(image_shifted),
            image_ref / jnp.linalg.norm(image_ref),
        )
    )

    # Peak gives the shift. can fit for subpixel accuracy if needed
    max_idx = jnp.unravel_index(jnp.argmax(abs_cross_corr), abs_cross_corr.shape)
    optimal_shift = coordinate_grid_in_angstroms[max_idx]

    return (
        jnp.max(abs_cross_corr),
        optimal_shift,
    )


def _compute_cross_correlation_image(image_shifted, image_ref):
    """Compute cross-correlation image between two input images.
    Using Fourier's convolution theorem.
    """
    image_shifted_fft = cxim.rfftn(image_shifted)
    image_ref_fft = cxim.rfftn(image_ref)
    cross_corr_fft = image_shifted_fft * image_ref_fft.conj()
    return cxim.irfftn(cross_corr_fft)



from ._forcefields import (
    AbstractForceField as AbstractForceField,
    compute_harmonic_steering_force as compute_harmonic_steering_force,
)
from ._langevin_dynamics import (
    OverdampedLangevinSampler as OverdampedLangevinSampler,
    ParallelSteeredOverdampedLangevinSampler as ParallelSteeredOverdampedLangevinSampler,
    SteeredOverdampedLangevinSampler as SteeredOverdampedLangevinSampler,
)
from ._molecular_dynamics import (
    EnsembleSteeredMDSimulator as EnsembleSteeredMDSimulator,
    SteeredMDSimulator as SteeredMDSimulator,
)
from .base_prior_projector import (
    AbstractEnsemblePriorProjector as AbstractEnsemblePriorProjector,
    AbstractPriorProjector as AbstractPriorProjector,
)



from pathlib import Path

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import optimistix as optx
from cryojax.constants import PengScatteringFactorParameters
from cryojax.io import read_atoms_from_pdb
from cryojax.jax_util import make_filter_spec
from jaxtyping import Array, Float, Int, PyTree


class Gaussian3D(eqx.Module, strict=True):
    positions: Float[Array, "n_beads 3"]
    amplitude: Float
    variance: Float
    shape: tuple[Int, Int, Int]
    voxel_size: Int
    n_gaussians_per_bead: Int

    def to_real_voxel_grid(
        self,
    ) -> Float[Array, "{self.shape[0]} {self.shape[1]} {self.shape[2]}"]:
        gmm_volume = self.to_gmm_volume()
        render_fn = cxs.GaussianMixtureRenderFn(self.shape, self.voxel_size)

        return render_fn(gmm_volume)

    def to_gmm_volume(self) -> cxs.GaussianMixtureVolume:
        ones = jnp.ones((self.positions.shape[0], self.n_gaussians_per_bead))

        gmm_volume = cxs.GaussianMixtureVolume(
            self.positions,
            amplitudes=self.amplitude * ones,
            variances=self.variance * ones,
        )
        return gmm_volume


def make_gmm_model_from_atomic_model(
    pdb_file: str | Path,
    box_size: int,
    voxel_size: float,
    *,
    fit_selection_string: str = 'name "C2"',
    init_amp: float = 6.0,
    init_var: float = 1.0,
    n_gaussians_per_bead: int = 1,
    atol: float = 1e-3,
    rtol: float = 1e-3,
    max_steps: int = 500,
) -> cxs.GaussianMixtureVolume:
    target_volume = _make_target_voxel_grid(pdb_file, box_size, voxel_size)

    init_gmm_model = _make_initial_gmm(
        pdb_file,
        box_size,
        voxel_size,
        selection_string=fit_selection_string,
        init_amp=init_amp,
        init_var=init_var,
        n_gaussians_per_bead=n_gaussians_per_bead,
    )

    return fit_gmm_model_to_voxel_grid(
        init_gmm_model, target_volume, atol=atol, rtol=rtol, max_steps=max_steps
    )


def fit_gmm_model_to_voxel_grid(
    init_gmm_model: Gaussian3D,
    target_volume: Float[Array, "z y x"],
    *,
    atol=1e-3,
    rtol=1e-3,
    max_steps=500,
) -> cxs.GaussianMixtureVolume:
    filter_spec = make_filter_spec(init_gmm_model, _where_gaussian3D_opt)

    gmm_model_opt, gmm_model_noopt = eqx.partition(init_gmm_model, filter_spec)
    y0, pytreedef = jax.tree.flatten(gmm_model_opt)

    sol = optx.least_squares(
        _compute_residues,
        y0=y0,
        args=(pytreedef, gmm_model_noopt, target_volume),
        solver=optx.GaussNewton(atol=atol, rtol=rtol),
        max_steps=max_steps,
        throw=True,
    )

    final_y = sol.value
    return eqx.combine(
        jax.tree.unflatten(pytreedef, final_y), gmm_model_noopt
    ).to_gmm_volume()


def _where_gaussian3D_opt(gaussian3d: Gaussian3D):
    return (
        gaussian3d.amplitude,
        gaussian3d.variance,
    )


def _make_target_voxel_grid(
    reference_pdb_file: str | Path, box_size: int, voxel_size: float
) -> Float[Array, "{box_size} {box_size} {box_size}"]:
    # read in atoms
    atom_positions, atomic_numbers, atomic_properties = read_atoms_from_pdb(
        reference_pdb_file,
        center=True,
        loads_properties=True,
        selection_string="not element H",
    )
    scattering_factor_parameters = PengScatteringFactorParameters(atomic_numbers)

    render_fn = cxs.GaussianMixtureRenderFn((box_size, box_size, box_size), voxel_size)
    gmm_volume = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions, scattering_factor_parameters, atomic_properties["b_factors"]
    )
    return render_fn(gmm_volume)


def _make_initial_gmm(
    reference_pdb_file: str | Path,
    box_size: int,
    voxel_size: float,
    *,
    selection_string: str = 'name "C2"',
    init_amp: float = 40.0,
    init_var: float = 1.0,
    n_gaussians_per_bead: int = 1,
) -> Gaussian3D:
    atom_positions, _ = read_atoms_from_pdb(
        reference_pdb_file, center=True, selection_string=selection_string
    )

    return Gaussian3D(
        positions=jnp.array(atom_positions),
        amplitude=init_amp,
        variance=init_var,
        shape=(box_size, box_size, box_size),
        voxel_size=voxel_size,
        n_gaussians_per_bead=n_gaussians_per_bead,
    )


@eqx.filter_jit
def _compute_residues(
    y: list[float], args: tuple[PyTree, Gaussian3D, Float[Array, "z y x"]]
) -> Float[Array, "z y x"]:
    pytreedef, gmm_model_noopt, target_volume = args
    gmm_volume = eqx.combine(
        jax.tree.unflatten(pytreedef, y), gmm_model_noopt
    ).to_real_voxel_grid()
    return gmm_volume - target_volume



import os
import pathlib
import shutil

import jax
import jax.numpy as jnp
import pytest
from cryospax import RelionParticleDataset, RelionParticleParameterFile

from cryojax_eo.dataset import (
    create_dataloader,
    make_relion_parameter_file,
    simulate_relion_dataset,
)
from cryojax_eo.internal import DatasetSimulatorConfig


@pytest.fixture
def sample_simulator_config(
    sample_path_to_pdb1,
    sample_path_to_pdb2,
):
    path_to_starfile = os.path.join(
        os.path.dirname(__file__), "outputs", "test_starfile.star"
    )
    path_to_relion_project = os.path.join(os.path.dirname(__file__), "outputs")
    return DatasetSimulatorConfig(
        number_of_images=10,
        data_sign="dark-on-light",
        pixel_size=0.8,
        box_size=32,
        pad_scale=1,
        voltage_in_kilovolts=300.0,
        offset_x_in_angstroms=0.0,
        offset_y_in_angstroms=0.0,
        defocus_in_angstroms=[100, 200],
        astigmatism_in_angstroms=[-1.0, 1.0],
        astigmatism_angle_in_degrees=[-1.0, 1.0],
        phase_shift=0.0,
        amplitude_contrast_ratio=0.1,
        spherical_aberration_in_mm=1e-16,
        ctf_scale_factor=1.0,
        envelope_b_factor=0.0,
        noise_snr=0.1,
        mask_radius=32 // 2,
        mask_rolloff_width=1.0,
        rng_seed=0,
        atomic_models_params=dict(
            path_to_atomic_models=[sample_path_to_pdb1, sample_path_to_pdb2],
            atomic_models_probabilities=[0.5, 0.5],
            loads_b_factors=True,
            atom_selection="not element H",
        ),
        path_to_relion_project=pathlib.Path(path_to_relion_project),
        path_to_starfile=pathlib.Path(path_to_starfile),
        images_per_file=10,
        batch_size_for_generation=10,
        overwrite=True,
    )


def test_make_relion_parameter_file(sample_simulator_config):
    parameter_file = make_relion_parameter_file(
        jax.random.key(0), sample_simulator_config
    )

    assert (
        len(parameter_file) == sample_simulator_config.number_of_images
    ), "Number of entries in parameter file different from requested"
    return


def test_simulate_relion_dataset(sample_simulator_config):
    relion_dataset = simulate_relion_dataset(sample_simulator_config)

    assert (
        len(relion_dataset) == sample_simulator_config.number_of_images
    ), "Number of images simulated different from requested"
    shutil.rmtree(relion_dataset.path_to_relion_project)
    return


# parametrize shuffle = True and False, and


@pytest.mark.parametrize("shuffle", [True, False])
def test_create_dataloader(
    shuffle, sample_path_to_starfile, sample_path_to_relion_project
):
    relion_stack_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            sample_path_to_starfile, options=dict(broadcasts_image_config=True)
        ),
        sample_path_to_relion_project,
    )

    per_particle_args = jnp.arange(0, len(relion_stack_dataset))

    dataloader = create_dataloader(
        relion_stack_dataset,
        batch_size=2,
        jax_prng_key=jax.random.key(0),
        shuffle=shuffle,
        per_particle_args=per_particle_args,
    )

    batch = next(iter(dataloader))

    assert "particle_stack" in batch, "Batch does not contain particle_stack"
    assert "per_particle_args" in batch, "Batch does not contain per_particle_args"
    assert batch["particle_stack"]["images"].shape[0] == 2, "Batch size is incorrect"
    assert (
        batch["per_particle_args"].shape[0] == 2
    ), "Batch size of per_particle_args is incorrect"
    return



from ._cross_corelation import (
    ModelToVolumeLikelihoodFn as ModelToVolumeLikelihoodFn,
    SteepestDescWalkerFlexibleFitting as SteepestDescWalkerFlexibleFitting,
)
from ._pipeline import FlexibleFittingPipeline as FlexibleFittingPipeline



import jax
import jax.numpy as jnp
import mdtraj
from cryospax import RelionParticleDataset, RelionParticleParameterFile

from cryojax_eo.dataset import create_dataloader
from cryojax_eo.ensemble_optimization import (
    ImagesToEnsembleLikelihoodFn,
    IterativeEnsembleLikelihoodOptimizer,
    MargGaussianWhiteLogLikelihoodFn,
    ProjGradDescWeightOptimizer,
)
from cryojax_eo.io import read_walkers_from_pdbs


def test_iterative_optimizer(
    sample_path_to_pdb1,
    sample_path_to_pdb2,
    sample_path_to_relion_project,
    sample_path_to_starfile,
):
    atom_indices = mdtraj.load(sample_path_to_pdb1).topology.select("not element H")
    walkers, amplitudes, variances = read_walkers_from_pdbs(
        [sample_path_to_pdb1, sample_path_to_pdb2]
    )
    walkers = walkers[:, atom_indices]
    variances = variances[atom_indices]
    amplitudes = amplitudes[atom_indices]

    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            sample_path_to_starfile, options=dict(broadcasts_image_config=True)
        ),
        sample_path_to_relion_project,
    )

    dataloader = create_dataloader(
        relion_dataset,
        batch_size=2,
        shuffle=True,
        per_particle_args=None,
        jax_prng_key=jax.random.key(0),
    )

    img_to_walker_likelihood_fn = MargGaussianWhiteLogLikelihoodFn(
        amplitudes=amplitudes,
        variances=variances,
        image_sign=1.0,
    )

    ensemble_likelihood_fn = ImagesToEnsembleLikelihoodFn(img_to_walker_likelihood_fn)

    optimizer = IterativeEnsembleLikelihoodOptimizer(
        step_size=1.0,
        n_steps=2,
        ensemble_likelihood_fn=ensemble_likelihood_fn,
        n_batches_per_step=2,
    )

    weights = jnp.array([0.5, 0.5])
    _, new_positions, new_weights = optimizer(walkers, weights, dataloader)

    assert (
        new_positions.shape == walkers.shape
    ), "Optimized positions have incorrect shape"
    assert new_weights.shape == weights.shape, "Optimized weights have incorrect shape"

    return


def test_weight_optimizer(
    sample_path_to_pdb1,
    sample_path_to_pdb2,
    sample_path_to_relion_project,
    sample_path_to_starfile,
):
    walkers, amplitudes, variances = read_walkers_from_pdbs(
        [sample_path_to_pdb1, sample_path_to_pdb2]
    )

    atom_indices = mdtraj.load(sample_path_to_pdb1).topology.select("not element H")
    walkers = walkers[:, atom_indices]
    variances = variances[atom_indices]
    amplitudes = amplitudes[atom_indices]

    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            sample_path_to_starfile, options=dict(broadcasts_image_config=True)
        ),
        sample_path_to_relion_project,
    )

    dataloader = create_dataloader(
        relion_dataset,
        batch_size=2,
        shuffle=True,
        per_particle_args=None,
        jax_prng_key=jax.random.key(0),
    )

    img_to_walker_likelihood_fn = MargGaussianWhiteLogLikelihoodFn(
        amplitudes=amplitudes,
        variances=variances,
        image_sign=1.0,
    )

    ensemble_likelihood_fn = ImagesToEnsembleLikelihoodFn(img_to_walker_likelihood_fn)

    optimizer = ProjGradDescWeightOptimizer(
        n_steps=2,
        ensemble_likelihood_fn=ensemble_likelihood_fn,
    )

    weights = jnp.array([0.5, 0.5])
    new_weights = optimizer(walkers, weights, dataloader)

    assert new_weights.shape == weights.shape, "Optimized weights have incorrect shape"

    return



from typing import Any

import cryojax.simulator as cxs
import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float

from ....simulator._dilated_mask import DilatedMask
from .common_functions import compute_optimal_scale_and_offset
from .make_model_utils import make_image_model_from_gmm


class AbstractImageToWalkerLogLikelihoodFn(eqx.Module, strict=True):
    amplitudes: eqx.AbstractVar[Float[Array, "n_atoms n_gaussians_per_atom"]]
    variances: eqx.AbstractVar[Float[Array, "n_atoms n_gaussians_per_atom"]]
    image_sign: eqx.AbstractVar[Float[Array, ""]]
    dilated_mask: eqx.AbstractVar[DilatedMask | None]

    def __call__(
        self,
        walker: Float[Array, "n_atoms 3"],
        image: Float[Array, "y x"],
        image_config: cxs.BasicImageConfig,
        pose: cxs.AbstractPose,
        transfer_theory: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float:
        raise NotImplementedError


class MargGaussianWhiteLogLikelihoodFn(AbstractImageToWalkerLogLikelihoodFn, strict=True):
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"]
    variances: Float[Array, "n_atoms n_gaussians_per_atom"]
    image_sign: Float[Array, ""]
    dilated_mask: DilatedMask | None

    def __init__(
        self,
        amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
        variances: Float[Array, "n_atoms n_gaussians_per_atom"],
        image_sign: Float[Array, ""] | float,
        dilated_mask: DilatedMask | None = None,
    ):
        assert (amplitudes > 0).all(), "Amplitudes must be positive."
        assert (variances > 0).all(), "Variances must be positive."

        self.variances = variances
        self.amplitudes = amplitudes
        self.image_sign = jnp.asarray(image_sign)
        self.dilated_mask = dilated_mask

    def __call__(
        self,
        walker: Float[Array, "n_atoms 3"],
        image: Float[Array, "y x"],
        image_config: cxs.BasicImageConfig,
        pose: cxs.AbstractPose,
        transfer_theory: cxs.ContrastTransferTheory,
        per_particle_args: Any,
    ) -> Float:
        return _likelihood_isotropic_gaussian_marginalized(
            walker=walker,
            amplitudes=self.amplitudes,
            variances=self.variances,
            image=image,
            image_config=image_config,
            pose=pose,
            transfer_theory=transfer_theory,
            dilated_mask=self.dilated_mask,
            image_sign=self.image_sign,
        )


class GaussianWhiteLogLikelihoodFn(AbstractImageToWalkerLogLikelihoodFn, strict=True):
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"]
    variances: Float[Array, "n_atoms n_gaussians_per_atom"]
    image_sign: Float[Array, ""]
    dilated_mask: DilatedMask | None

    def __init__(
        self,
        amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
        variances: Float[Array, "n_atoms n_gaussians_per_atom"],
        image_sign: Float[Array, ""],
        dilated_mask: DilatedMask | None = None,
    ):
        assert (amplitudes > 0).all(), "Amplitudes must be positive."
        assert (variances > 0).all(), "Variances must be positive."

        self.variances = variances
        self.amplitudes = amplitudes
        self.image_sign = image_sign
        self.dilated_mask = dilated_mask

    def __call__(
        self,
        walker: Float[Array, "n_atoms 3"],
        image: Float[Array, "y x"],
        image_config: cxs.BasicImageConfig,
        pose: cxs.AbstractPose,
        transfer_theory: cxs.ContrastTransferTheory,
        per_particle_args: Float[Array, ""],
    ) -> Float:
        return _likelihood_gaussian_white_noise(
            walker=walker,
            amplitudes=self.amplitudes,
            variances=self.variances,
            image=image,
            image_config=image_config,
            pose=pose,
            transfer_theory=transfer_theory,
            dilated_mask=self.dilated_mask,
            image_sign=self.image_sign,
            per_particle_args=per_particle_args,
        )


def _likelihood_gaussian_white_noise(
    walker: Float[Array, "n_atoms 3"],
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
    variances: Float[Array, "n_atoms n_gaussians_per_atom"],
    image: Float[Array, "y x"],
    image_config: cxs.BasicImageConfig,
    pose: cxs.AbstractPose,
    transfer_theory: cxs.ContrastTransferTheory,
    dilated_mask: DilatedMask | None = None,
    image_sign: Float[Array, ""] = jnp.array(1.0),
    per_particle_args: Float[Array, ""] = jnp.array(1.0),
) -> Float:
    """
    Compute the likelihood of a walker given a Relion stack using an isotropic Gaussian
    likelihood function.

    **Arguments:**
    - `walker`: A `walker` that is, a point cloud representing an atomic model.
    - `relion_stack`: A cryojax `ParticleStack` object.
    - `amplitudes`: The amplitudes for the GMM atomic volume representation.
    - `variances`: The variances for the GMM atomic volume representation.
    - `dilated_mask`: An optional dilated mask to apply to the computed image.
    - `image_sign`: For this particular function the constant argument
        is the sign of the observed image. For typical Relion stacks this is -1.0.
        For data generated with cryoJAX this is 1.0.
    - `per_particle_args`: The noise variance for the likelihood function.

    **Returns:**
    - The log likelihood of the walker given the Relion stack.

    """

    noise_variance = per_particle_args

    image_model = make_image_model_from_gmm(
        walker, amplitudes, variances, image_config, pose, transfer_theory
    )
    computed_image = image_model.simulate()
    observed_image = jnp.asarray(image)

    if dilated_mask is not None:
        mask2d = dilated_mask.project(pose)
    else:
        mask2d = jnp.ones_like(computed_image)

    computed_image = computed_image * mask2d
    observed_image = image_sign * observed_image * mask2d

    scale, offset = compute_optimal_scale_and_offset(computed_image, observed_image)

    return -jnp.sum((scale * computed_image - observed_image + offset) ** 2) / (
        2 * noise_variance
    )


def _likelihood_isotropic_gaussian_marginalized(
    walker: Float[Array, "n_atoms 3"],
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
    variances: Float[Array, "n_atoms n_gaussians_per_atom"],
    image: Float[Array, "y x"],
    image_config: cxs.BasicImageConfig,
    pose: cxs.AbstractPose,
    transfer_theory: cxs.ContrastTransferTheory,
    dilated_mask: DilatedMask | None = None,
    image_sign: Float[Array, ""] = jnp.array(1.0),
    per_particle_args: None = None,
) -> Float:
    """
    Compute the marginalized likelihood of a walker given a Relion stack using an
    isotropic Gaussian likelihood function where the variance has been marginalized.
    This is useful when the variance is not known or is not fixed.
    """
    assert (
        per_particle_args is None
    ), "per_particle_args is not used in this function and should be None."

    image_model = make_image_model_from_gmm(
        walker, amplitudes, variances, image_config, pose, transfer_theory
    )
    computed_image = image_model.simulate()
    observed_image = jnp.asarray(image)

    if dilated_mask is not None:
        mask2d = dilated_mask.project(pose)
    else:
        mask2d = jnp.ones_like(computed_image)

    computed_image = computed_image * mask2d
    observed_image = image_sign * observed_image * mask2d

    scale, offset = compute_optimal_scale_and_offset(computed_image, observed_image)
    n_pixels = computed_image.size
    loss = -n_pixels * jnp.log(
        jnp.linalg.norm(scale * computed_image - observed_image + offset)
    )
    return loss



from typing import TypeVar
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, Int, PRNGKeyArray

from .._forcefields.base_forcefield import AbstractForceField
from .._forcefields.biasing_forces import compute_harmonic_steering_force
from ..base_prior_projector import AbstractPriorProjector


EnergyFuncArgs = TypeVar("EnergyFuncArgs")


class OverdampedLangevinSampler(AbstractPriorProjector):
    """
    Overdamped Langevin sampler for Overdamped Langevin dynamics.
    """

    n_steps: Int
    step_size: Float
    forcefield: AbstractForceField

    def __init__(
        self,
        n_steps: Int,
        step_size: Float,
        forcefield: AbstractForceField,
    ):
        """
        **Arguments:**
        - `n_steps`: Number of steps to take.
        - `step_size`: Step size for Overdamped Langevin dynamics.
        - `forcefield`: Force field to use for Overdamped Langevin dynamics.
        """
        self.n_steps = n_steps
        self.step_size = step_size
        self.forcefield = forcefield

    @override
    def initialize(
        self, init_state: tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]]
    ) -> tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]]:
        """
        Initialize the sampler with the initial state. For this sampler, the initial state
        is the indentity, and its purpsoe is simply to validate the type of the input.

        **Arguments:**
        init_state: Initial positions of the atoms or walkers, and initial key.

        **Returns:**
        Initial state for the sampler.
        """
        key, initial_walker = init_state

        assert (
            initial_walker.ndim == 2 and initial_walker.shape[1] == 3
        ), "initial_walker must be a 2D array with shape (n_atoms, 3)"
        assert jnp.isdtype(
            initial_walker.dtype, "real floating"
        ), "initial_walker must be a real floating point array"

        return init_state

    def __call__(
        self,
        state: tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]],
        *,
        outputs_trajectory: bool = False,
    ) -> tuple[
        Float[Array, "n_atoms 3"],
        tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]],
    ]:
        """
        Sample a trajectory from the initial atom positions.

        **Arguments:**
        key: JAX random key for generating random numbers.
        initial_walkers: Initial positions of the atoms.
        energy_fn_args: Arguments for the energy function.

        **Returns:**
        trajectory: Overdamped Langevin Dynamics Trajectory.
        """

        key, initial_walkers = state
        key, trajectory = _run_overdamped_langevin(
            key,
            initial_walkers,
            self.n_steps,
            self.step_size,
            self.forcefield,
            outputs_trajectory=outputs_trajectory,
        )
        return trajectory, (key, trajectory[-1])


class SteeredOverdampedLangevinSampler(AbstractPriorProjector):
    """
    Overdamped Langevin sampler for Overdamped Langevin dynamics.
    """

    n_steps: Int
    step_size: Float
    forcefield: AbstractForceField
    biasing_force_constant: Float

    def __init__(
        self,
        n_steps: Int,
        step_size: Float,
        forcefield: AbstractForceField,
        biasing_force_constant: Float,
    ):
        """
        **Arguments:**
        - `n_steps`: Number of steps to take.
        - `step_size`: Step size for Overdamped Langevin dynamics.
        - `forcefield`: Force field to use for Overdamped Langevin dynamics.
        - `biasing_force_constant`: Force constant for the biasing volume.
        """
        self.n_steps = n_steps
        self.step_size = step_size
        self.forcefield = forcefield
        self.biasing_force_constant = biasing_force_constant

    @override
    def initialize(
        self, init_state: tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]]
    ) -> tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]]:
        """
        Initialize the sampler with the initial state. For this sampler, the initial state
        is the indentity, and its purpsoe is simply to validate the type of the input.

        **Arguments:**
        init_state: Initial positions of the atoms or walkers, and initial key.

        **Returns:**
        Initial state for the sampler.
        """
        key, initial_walker = init_state

        assert (
            initial_walker.ndim == 2 and initial_walker.shape[1] == 3
        ), "initial_walker must be a 2D array with shape (n_atoms, 3)"
        assert jnp.isdtype(
            initial_walker.dtype, "real floating"
        ), "initial_walker must be a real floating point array"

        return init_state

    def __call__(
        self,
        ref_walkers: Float[Array, "n_atoms 3"],
        state: tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]],
        *,
        outputs_trajectory: bool = False,
    ) -> tuple[
        Float[Array, "n_atoms 3"],
        tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]],
    ]:
        """
        Sample a trajectory from the initial atom positions.

        **Arguments:**
        key: JAX random key for generating random numbers.
        initial_walkers: Initial positions of the atoms.
        energy_fn_args: Arguments for the energy function.

        **Returns:**
        trajectory: Overdamped Langevin Dynamics Trajectory.
        """

        key, initial_walkers = state
        key, walkers = _run_steered_overdamped_langevin(
            key,
            initial_walkers,
            ref_walkers,
            self.n_steps,
            self.step_size,
            self.forcefield,
            self.biasing_force_constant,
            outputs_trajectory=outputs_trajectory,
        )
        return walkers, (key, walkers)


class ParallelSteeredOverdampedLangevinSampler(AbstractPriorProjector):
    """
    Parallel Overdamped Langevin sampler for Overdamped Langevin dynamics.
    """

    n_steps: Int
    step_size: Float
    forcefield: AbstractForceField
    biasing_force_constant: Float

    def __init__(
        self,
        n_steps: Int,
        step_size: Float,
        forcefield: AbstractForceField,
        biasing_force_constant: Float,
    ):
        """
        **Arguments:**
        - `n_steps`: Number of steps to take.
        - `step_size`: Step size for Overdamped Langevin dynamics.
        - `forcefield`: Force field to use for Overdamped Langevin dynamics.
        - `biasing_force_constant`: Force constant for the biasing potential.
        """
        self.n_steps = n_steps
        self.step_size = step_size
        self.forcefield = forcefield
        self.biasing_force_constant = biasing_force_constant

    @override
    def initialize(
        self, init_state: tuple[PRNGKeyArray, Float[Array, "n_walkers n_atoms 3"]]
    ) -> tuple[PRNGKeyArray, Float[Array, "n_atoms 3"]]:
        """
        Initialize the sampler with the initial state. For this sampler, the initial state
        is the indentity, and its purpsoe is simply to validate the type of the input.

        **Arguments:**
        init_state: Initial positions of the atoms or walkers, and initial key.

        **Returns:**
        Initial state for the sampler.
        """
        key, initial_walker = init_state

        assert (
            initial_walker.ndim == 3 and initial_walker.shape[2] == 3
        ), "initial_walker must be a 2D array with shape (n_atoms, 3)"
        assert jnp.isdtype(
            initial_walker.dtype, "real floating"
        ), "initial_walker must be a real floating point array"

        return init_state

    def __call__(
        self,
        ref_walkers: Float[Array, "n_walkers n_atoms 3"],
        state: tuple[PRNGKeyArray, Float[Array, "n_walkers n_atoms 3"]],
        *,
        outputs_trajectory: bool = False,
    ) -> tuple[
        Float[Array, "n_walkers n_atoms 3"],
        tuple[PRNGKeyArray, Float[Array, "n_walkers n_atoms 3"]],
    ]:
        """
        Sample a trajectory from the initial atom positions.

        **Arguments:**
        key: JAX random key for generating random numbers.
        initial_walkers: Initial positions of the atoms.
        energy_fn_args: Arguments for the energy function.

        **Returns:**
        trajectory: Overdamped Langevin Dynamics Trajectory.
        """

        key, initial_walkers = state
        key, walkers = _run_steered_overdamped_langevin_parallel(
            key,
            initial_walkers,
            ref_walkers,
            self.n_steps,
            self.step_size,
            self.forcefield,
            self.biasing_force_constant,
            outputs_trajectory=outputs_trajectory,
        )
        return walkers, (key, walkers)


###### Steered Parallel ######


@eqx.filter_jit
def _take_overdamped_langevin_step(
    key: PRNGKeyArray,
    atom_positions: Float[Array, "n_atoms 3"],
    forcefield: AbstractForceField,
    step_size: Float,
) -> Float[Array, "n_atoms 3"]:
    gradient = jax.grad(lambda x: forcefield(x))(atom_positions)

    new_positions = (
        atom_positions
        - step_size * gradient
        + jnp.sqrt(2 * step_size) * jax.random.normal(key, shape=atom_positions.shape)
    )
    return new_positions


def _run_overdamped_langevin(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    *,
    outputs_trajectory: bool = False,
):
    if outputs_trajectory:
        return _run_overdamped_langevin_scan(
            key,
            initial_walker,
            n_steps,
            step_size,
            forcefield,
        )
    else:
        return _run_overdamped_langevin_foriloop(
            key,
            initial_walker,
            n_steps,
            step_size,
            forcefield,
        )


@eqx.filter_jit
def _run_overdamped_langevin_foriloop(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
):
    def _step_func(i, val):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_overdamped_langevin_step(
            subkey,
            old_positions,
            forcefield,
            step_size,
        )
        return (key, new_positions)

    return jax.lax.fori_loop(
        lower=0,
        upper=n_steps,
        body_fun=_step_func,
        init_val=(key, initial_walker),
    )


@eqx.filter_jit
def _run_overdamped_langevin_scan(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
):
    def _step_func(val, i):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_overdamped_langevin_step(
            subkey,
            old_positions,
            forcefield,
            step_size,
        )
        return (key, new_positions), new_positions

    return jax.lax.scan(f=_step_func, init=(key, initial_walker), xs=jnp.arange(n_steps))


###### Steered Single ######


@eqx.filter_jit
def _take_steered_overdamped_langevin_step(
    key: PRNGKeyArray,
    atom_positions: Float[Array, "n_atoms 3"],
    ref_walkers: Float[Array, "n_atoms 3"],
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
    step_size: Float,
) -> Float[Array, "n_atoms 3"]:
    energy_gradient = jax.grad(lambda x: forcefield(x))(atom_positions)

    biasing_gradient = -compute_harmonic_steering_force(
        atom_positions, ref_walkers, biasing_force_constant
    )  # it's negative since it's a force F = -grad E

    gradient = energy_gradient + biasing_gradient
    # jax.debug.print("{gradient}", gradient=gradient)
    new_positions = (
        atom_positions
        - step_size * gradient
        + jnp.sqrt(2 * step_size) * jax.random.normal(key, shape=atom_positions.shape)
    )
    return new_positions


@eqx.filter_jit
def _run_steered_overdamped_langevin(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    reference_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
    *,
    outputs_trajectory: bool = False,
):
    if outputs_trajectory:
        return _run_steered_overdamped_langevin_scan(
            key,
            initial_walker,
            reference_walker,
            n_steps,
            step_size,
            forcefield,
            biasing_force_constant,
        )
    else:
        return _run_steered_overdamped_langevin_foriloop(
            key,
            initial_walker,
            reference_walker,
            n_steps,
            step_size,
            forcefield,
            biasing_force_constant,
        )


@eqx.filter_jit
def _run_steered_overdamped_langevin_foriloop(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    reference_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
):
    def _step_func(i, val):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_steered_overdamped_langevin_step(
            subkey,
            old_positions,
            reference_walker,
            forcefield,
            biasing_force_constant,
            step_size,
        )
        return (key, new_positions)

    return jax.lax.fori_loop(
        lower=0,
        upper=n_steps,
        body_fun=_step_func,
        init_val=(key, initial_walker),
    )


@eqx.filter_jit
def _run_steered_overdamped_langevin_scan(
    key: PRNGKeyArray,
    initial_walker: Float[Array, "n_atoms 3"],
    reference_walker: Float[Array, "n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
):
    def _step_func(val, i):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_steered_overdamped_langevin_step(
            subkey,
            old_positions,
            reference_walker,
            forcefield,
            biasing_force_constant,
            step_size,
        )
        return (key, new_positions), new_positions

    return jax.lax.scan(f=_step_func, init=(key, initial_walker), xs=jnp.arange(n_steps))


###### Steered Parallel ######


@eqx.filter_jit
def _take_steered_overdamped_langevin_step_parallel(
    key: PRNGKeyArray,
    atom_positions: Float[Array, "n_walkers n_atoms 3"],
    ref_walkers: Float[Array, "n_walkers n_atoms 3"],
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
    step_size: Float,
) -> Float[Array, "n_walkers n_atoms 3"]:
    energy_gradient = eqx.filter_vmap(jax.grad(lambda x: forcefield(x)))(atom_positions)

    biasing_gradient = -compute_harmonic_steering_force(
        atom_positions, ref_walkers, biasing_force_constant
    )

    gradient = energy_gradient + biasing_gradient

    new_positions = (
        atom_positions
        - step_size * gradient
        + jnp.sqrt(2 * step_size) * jax.random.normal(key, shape=atom_positions.shape)
    )

    return new_positions


def _run_steered_overdamped_langevin_parallel(
    key: PRNGKeyArray,
    initial_walkers: Float[Array, "n_walkers n_atoms 3"],
    reference_walkers: Float[Array, "n_walkers n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
    *,
    outputs_trajectory: bool = False,
):
    if outputs_trajectory:
        return _run_steered_overdamped_langevin_parallel_scan(
            key,
            initial_walkers,
            reference_walkers,
            n_steps,
            step_size,
            forcefield,
            biasing_force_constant,
        )
    else:
        return _run_steered_overdamped_langevin_parallel_foriloop(
            key,
            initial_walkers,
            reference_walkers,
            n_steps,
            step_size,
            forcefield,
            biasing_force_constant,
        )


@eqx.filter_jit
def _run_steered_overdamped_langevin_parallel_foriloop(
    key: PRNGKeyArray,
    initial_walkers: Float[Array, "n_walkers n_atoms 3"],
    reference_walkers: Float[Array, "n_walkers n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
):
    def _step_func(i, val):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_steered_overdamped_langevin_step_parallel(
            subkey,
            old_positions,
            reference_walkers,
            forcefield,
            biasing_force_constant,
            step_size,
        )
        return (key, new_positions)

    return jax.lax.fori_loop(
        lower=0,
        upper=n_steps,
        body_fun=_step_func,
        init_val=(key, initial_walkers),
    )


@eqx.filter_jit
def _run_steered_overdamped_langevin_parallel_scan(
    key: PRNGKeyArray,
    initial_walkers: Float[Array, "n_walkers n_atoms 3"],
    reference_walkers: Float[Array, "n_walkers n_atoms 3"],
    n_steps: int,
    step_size: Float,
    forcefield: AbstractForceField,
    biasing_force_constant: Float,
):
    def _step_func(i, val):
        key, old_positions = val
        key, subkey = jax.random.split(key)
        new_positions = _take_steered_overdamped_langevin_step_parallel(
            subkey,
            old_positions,
            reference_walkers,
            forcefield,
            biasing_force_constant,
            step_size,
        )
        return (key, new_positions), new_positions

    return jax.lax.scan(f=_step_func, init=(key, initial_walkers), xs=jnp.arange(n_steps))



#!/usr/bin/env python3
import argparse
import datetime
import logging
import os
from pathlib import Path

import jax
import jax.numpy as jnp
import mdtraj
import mrcfile
import numpy as np
import optax
import yaml
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import fourier_crop_to_shape
from cryospax import (
    RelionParticleDataset,
    RelionParticleParameterFile,
)

import cryojax_eo as cxeo
from cryojax_eo.ensemble_optimization import (
    EnsembleOptimizationPipeline,
    EnsembleSteeredMDSimulator,
    ImagesToEnsembleLikelihoodFn,
    IterativeEnsembleLikelihoodOptimizer,
    MargGaussianWhiteLogLikelihoodFn,
    SteeredMDSimulator,
)
from cryojax_eo.internal import EnsOptMDConfig
from cryojax_eo.io import read_walkers_from_pdbs
from cryojax_eo.simulator import DilatedMask
from cryojax_eo.utils import ModelToVolumeAligner


def add_args(parser):
    parser.add_argument(
        "--config", type=str, help="Path to the config (yaml) file", required=True
    )
    return parser


def mkbasedir(out):
    if not os.path.exists(out):
        try:
            os.makedirs(out)
        except (FileExistsError, PermissionError):
            raise ValueError("Output path does not exist and cannot be created.")
    return


def warnexists(out):
    if os.path.exists(out):
        Warning(f"Warning: {out} already exists. Overwriting.")
    return


def _make_atom_list(atom_selection, topology) -> np.ndarray:
    suffix = Path(atom_selection).suffix
    if suffix in [".txt", ".npy"]:
        atom_list = np.loadtxt(atom_selection, dtype=int)
    else:
        atom_list = topology.select(atom_selection)
    return np.array(atom_list)


def run_ensemble_optimization_with_md(ensemble_opt_config: EnsOptMDConfig):
    config = dict(ensemble_opt_config.model_dump())

    # Load the initial walkers and reference structure

    logging.debug("Loading atomic models...")
    initial_walkers, variances, amplitudes = read_walkers_from_pdbs(
        config["path_to_atomic_models"],
        loads_b_factors=config["loads_b_factors"],
    )

    ref_structure = mdtraj.load(
        config["alignment_params"]["path_to_prealigned_atomic_model"]
    )
    ref_structure = ref_structure.center_coordinates(mass_weighted=True)

    atom_list = _make_atom_list(config["atom_selection"], ref_structure.topology)
    variances = variances[atom_list]
    amplitudes = amplitudes[atom_list]

    logging.debug("Atomic models loaded.")

    logging.debug("Loading experimental data...")
    # Load experimental data: images, mask, and consensus volume
    stack_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            path_to_starfile=config["data_params"]["path_to_starfile"],
            options=dict(
                loads_envelope=config["data_params"]["loads_envelope"],
                broadcasts_image_config=True,
            ),
        ),
        path_to_relion_project=config["data_params"]["path_to_relion_project"],
    )

    key = jax.random.PRNGKey(config["rng_seed"])
    key_data, key_pipeline = jax.random.split(key)

    dataloader = cxeo.dataset.create_dataloader(
        stack_dataset,
        batch_size=config["likelihood_optimizer_params"]["batch_size"],
        shuffle=True,
        drop_last=False,
        jax_prng_key=key_data,
    )
    logging.debug("Experimental data loaded.")

    if config["data_params"]["path_to_volumetric_mask"] is not None:
        logging.debug("Loading volumetric mask...")
        mask = jnp.asarray(
            mrcfile.open(
                config["data_params"]["path_to_volumetric_mask"],
                mode="r",
            ).data
        ).copy()
        dilated_mask = DilatedMask(mask, stack_dataset[0]["parameters"]["image_config"])  # type: ignore
        logging.debug("Volumetric mask loaded.")

    else:
        mask = None
        dilated_mask = None

    if config["alignment_params"]["path_to_reference_volume"] is not None:
        logging.debug("Loading consensus volume for alignment...")
        volume_for_alignment, voxel_size = read_array_from_mrc(
            config["alignment_params"]["path_to_reference_volume"],
            loads_grid_spacing=True,
        )

        if mask is not None:
            volume_for_alignment *= mask

        if config["alignment_params"]["reference_volume_voxel_size"] is not None:
            voxel_size = config["alignment_params"]["reference_volume_voxel_size"]

        box_size_ds = int(config["alignment_params"]["downsample_box_size"])

        voxel_size = voxel_size * volume_for_alignment.shape[0] / box_size_ds
        volume_for_alignment = fourier_crop_to_shape(
            volume_for_alignment, (box_size_ds, box_size_ds, box_size_ds)
        )

        model_aligner = ModelToVolumeAligner(volume_for_alignment, voxel_size)
        logging.debug("Consensus volume loaded.")
    else:
        model_aligner = None

    # Construct prior projector
    projector_list = []

    for i in range(initial_walkers.shape[0]):
        projector_list.append(
            SteeredMDSimulator(
                path_to_initial_pdb=config["path_to_atomic_models"][i],
                n_steps=config["projector_params"]["n_steps"],
                restrain_atom_list=atom_list.tolist(),
                parameters_for_md={
                    "platform": config["projector_params"]["platform"],
                    "properties": config["projector_params"]["platform_properties"],
                },
                base_state_file_path=os.path.join(
                    config["path_to_output"], f"states_proj_{i}/state_"
                ),
            )
        )
    md_projector = EnsembleSteeredMDSimulator(projector_list)

    # Construct likelihood optimizer
    image_sign = -1.0 if config["data_params"]["data_sign"] == "dark-on-light" else 1.0

    img_to_walker_log_likelihood_fn = MargGaussianWhiteLogLikelihoodFn(
        amplitudes,
        variances,
        image_sign=jnp.asarray(image_sign),
        dilated_mask=dilated_mask,
    )
    ensemble_likelihood_fn = ImagesToEnsembleLikelihoodFn(
        img_to_walker_log_likelihood_fn, n_walkers_in_parallel=1, n_images_in_parallel=50
    )
    if config["likelihood_optimizer_params"]["estimates_pose"]:
        raise NotImplementedError(
            "Pose estimation inside the MD ensemble"
            " optimization pipeline is not yet implemented."
        )

    likelihood_optimizer = IterativeEnsembleLikelihoodOptimizer(
        step_size=config["likelihood_optimizer_params"]["step_size"],
        n_steps=config["likelihood_optimizer_params"]["n_steps"],
        n_batches_per_step=config["likelihood_optimizer_params"]["n_batches_per_step"],
        ensemble_likelihood_fn=ensemble_likelihood_fn,
        pose_search=None,
    )

    runs_postprocessing = True if initial_walkers.shape[0] > 1 else False

    # Construct the ensemble optimization pipeline
    ensemble_refinement_pipeline = EnsembleOptimizationPipeline(
        prior_projector=md_projector,
        likelihood_optimizer=likelihood_optimizer,
        n_steps=config["n_steps"],
        prealigned_structure=ref_structure,
        atom_indices_for_opt=jnp.asarray(atom_list, dtype=int),
        model_to_volume_aligner=model_aligner,
        runs_postprocessing=runs_postprocessing,
    )

    # Running the optimization

    if isinstance(config["projector_params"]["bias_constant_in_kjpermol"], float):
        bias_constant_scheduler = optax.constant_schedule(
            config["projector_params"]["bias_constant_in_kjpermol"]
        )
    elif (
        isinstance(config["projector_params"]["bias_constant_in_kjpermol"], list)
        and len(config["projector_params"]["bias_constant_in_kjpermol"]) == 2
    ):
        bias_constant_scheduler = optax.linear_schedule(
            init_value=config["projector_params"]["bias_constant_in_kjpermol"][0],
            end_value=config["projector_params"]["bias_constant_in_kjpermol"][1],
            transition_steps=config["n_steps"],
        )
    else:
        raise ValueError(
            "bias_constant_in_kjpermol must be a float or a list of two floats."
        )

    initial_weights = jnp.array(config["likelihood_optimizer_params"]["initial_weights"])

    walkers, weights = ensemble_refinement_pipeline.run(
        key=key_pipeline,
        initial_walkers=initial_walkers,
        initial_weights=initial_weights,
        dataloader=dataloader,
        bias_constant_scheduler=bias_constant_scheduler,
        output_directory=config["path_to_output"],
        initial_state_for_projector=config["projector_params"]["path_to_initial_states"],
    )

    jnp.savez(
        os.path.join(config["path_to_output"], "final_ensemble.npz"),
        walkers=walkers,
        weights=weights,
    )

    return walkers, weights


def main(args):
    with open(args.config) as f:
        config_dict = yaml.safe_load(f)
        config = EnsOptMDConfig(**config_dict)

    warnexists(config.path_to_output)
    mkbasedir(config.path_to_output)

    logger = logging.getLogger()
    logger.handlers.clear()

    logger_fname = datetime.datetime.now().strftime("%Y-%m-%d")
    logger_fname = os.path.join(config.path_to_output, logger_fname + ".log")

    fhandler = logging.FileHandler(filename=logger_fname, mode="a")
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fhandler.setFormatter(formatter)
    logger.addHandler(fhandler)
    logger.setLevel(logging.INFO)

    config_fname = os.path.basename(args.config)
    with open(os.path.join(config.path_to_output, config_fname), "w") as f:
        yaml.dump(config_dict, f, default_flow_style=False)

    logging.info(
        f"A copy of the used config file has been written "
        f"to {os.path.join(config.path_to_output, config_fname)}"
    )

    logging.info("Running ensemble optimization...")
    _, _ = run_ensemble_optimization_with_md(config)
    logging.info("Ensemble optimization complete.")

    return


def main_cli():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=yaml.dump(EnsOptMDConfig.model_json_schema(), indent=4),
    )
    main(add_args(parser).parse_args())


if __name__ == "__main__":
    main_cli()



from ._loss_functions import (
    AbstractImageToWalkerLogLikelihoodFn as AbstractImageToWalkerLogLikelihoodFn,
    GaussianWhiteLogLikelihoodFn as GaussianWhiteLogLikelihoodFn,
    ImagesToEnsembleLikelihoodFn as ImagesToEnsembleLikelihoodFn,
    MargGaussianWhiteLogLikelihoodFn as MargGaussianWhiteLogLikelihoodFn,
    compute_optimal_scale_and_offset as compute_optimal_scale_and_offset,
    make_image_model_from_gmm as make_image_model_from_gmm,
)
from ._walker_optimizers import (
    IterativeEnsembleLikelihoodOptimizer as IterativeEnsembleLikelihoodOptimizer,
)
from ._weights_optimizer import ProjGradDescWeightOptimizer as ProjGradDescWeightOptimizer
from .base_optimizer import (
    AbstractEnsembleParameterOptimizer as AbstractEnsembleParameterOptimizer,
)



import cryojax.simulator as cxs
import jax
import numpy as np
import pytest
from cryospax import RelionParticleDataset, RelionParticleParameterFile

from cryojax_eo.ensemble_optimization import (
    GaussianWhiteLogLikelihoodFn,
    MargGaussianWhiteLogLikelihoodFn,  # done
    compute_optimal_scale_and_offset,  # done
    make_image_model_from_gmm,  # done
)
from cryojax_eo.io import read_walkers_from_pdbs
from cryojax_eo.simulator import DilatedMask


@pytest.fixture
def sample_relion_stack():
    return {
        "parameters": {
            "pose": cxs.EulerAnglePose(),
            "transfer_theory": cxs.ContrastTransferTheory(ctf=cxs.AstigmaticCTF()),
            "image_config": cxs.BasicImageConfig(
                shape=(32, 32), pixel_size=0.2, voltage_in_kilovolts=300.0
            ),
        },
        "images": jax.random.normal(jax.random.key(0), shape=(32, 32)),
    }


@pytest.fixture
def simple_volume_mask(sample_path_to_starfile):
    image_config = RelionParticleParameterFile(sample_path_to_starfile)[0]["image_config"]
    volume_shape = (image_config.shape[0], image_config.shape[0], image_config.shape[0])

    voxel_grid = jax.random.randint(jax.random.key(0), volume_shape, minval=0, maxval=2)

    return DilatedMask(voxel_grid, image_config)


def test_make_image_model_from_gmm(sample_path_to_pdb1, sample_relion_stack):
    walkers, amplitudes, variances = read_walkers_from_pdbs([sample_path_to_pdb1])

    make_image_model_from_gmm(
        walkers[0],
        amplitudes,
        variances,
        sample_relion_stack["parameters"]["image_config"],
        sample_relion_stack["parameters"]["pose"],
        sample_relion_stack["parameters"]["transfer_theory"],
    )
    return


def test_compute_scale_and_offset():
    scale = 2.0
    offset = -3.0

    image = jax.random.normal(jax.random.key(0), shape=(32, 32))
    image_transf = scale * image + offset

    computed_scale, computed_offset = compute_optimal_scale_and_offset(
        image, image_transf
    )

    np.testing.assert_allclose(
        computed_scale, scale, err_msg="Computed scale does not match true scale"
    )
    np.testing.assert_allclose(
        computed_offset, offset, err_msg="Computed offset does not match true offset"
    )
    return


@pytest.mark.parametrize("use_dilated_mask", [True, False])
@pytest.mark.parametrize(
    "image_to_walker_likelihood_fn",
    [MargGaussianWhiteLogLikelihoodFn, GaussianWhiteLogLikelihoodFn],
)
def test_likelihood_fn(
    sample_path_to_pdb1,
    sample_path_to_starfile,
    sample_path_to_relion_project,
    image_to_walker_likelihood_fn,
    use_dilated_mask,
    simple_volume_mask,
):
    walkers, amplitudes, variances = read_walkers_from_pdbs([sample_path_to_pdb1])

    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            path_to_starfile=sample_path_to_starfile,
            options=dict(broadcasts_image_config=True),
        ),
        path_to_relion_project=sample_path_to_relion_project,
        mode="r",
    )

    if use_dilated_mask:
        dilated_mask = simple_volume_mask
    else:
        dilated_mask = None

    img_to_walker_likelihood_fn = image_to_walker_likelihood_fn(
        amplitudes=amplitudes,
        variances=variances,
        image_sign=1.0,
        dilated_mask=dilated_mask,
    )
    stack = relion_dataset[0]
    img_to_walker_likelihood_fn(
        walker=walkers[0],
        image=stack["images"],
        image_config=stack["parameters"]["image_config"],
        pose=stack["parameters"]["pose"],
        transfer_theory=stack["parameters"]["transfer_theory"],
        per_particle_args=1.0,
    )
    return


# def test_likelihood_isotropic_gaussian():
#     random_image = jax.random.normal(jax.random.PRNGKey(0), (10, 10))
#     noise_variance = 1
#     assert jnp.isclose(
#         likelihood_isotropic_gaussian(random_image, random_image, noise_variance), 0.0
#     )

#     constant_image = jnp.ones((10, 10))
#     assert jnp.isnan(
#         likelihood_isotropic_gaussian(constant_image, random_image, noise_variance)
#     )

#     linear_image = jnp.linspace(0, 1, 100).reshape(10, 10)
#     random_scale = jax.random.uniform(jax.random.PRNGKey(1), (1,))
#     random_bias = jax.random.uniform(jax.random.PRNGKey(2), (1,))
#     assert jnp.isclose(
#         likelihood_isotropic_gaussian(
#             linear_image, random_scale * linear_image + random_bias, noise_variance
#         ),
#         0.0,
#     )

"""
def test_compute_likelihood_matrix(
    sample_path_to_starfile, sample_path_to_relion_project
):
    key = jax.random.key(0)
    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            path_to_starfile=sample_path_to_starfile,
        ),
        path_to_relion_project=sample_path_to_relion_project,
        mode="r",
    )

    n_walkers = 2
    n_atoms = int(jax.random.randint(key, (1,), 10, 20)[0])
    n_gaussians_per_atom = 5
    ensemble_walkers = jax.random.normal(key, (n_walkers, n_atoms, 3))
    amplitudes = jax.random.normal(key, (n_walkers, n_atoms, n_gaussians_per_atom))
    variances = jax.random.normal(key, (n_walkers, n_atoms, n_gaussians_per_atom)) ** 2

    # test of likelihood_isotropic_gaussian
    image_to_walker_log_likelihood_fn = likelihood_isotropic_gaussian
    per_particle_args_noise_variance = 1.0

    n_particles = 5
    likelihood_matrix = compute_likelihood_matrix(
        ensemble_walkers,
        relion_dataset[:n_particles],
        amplitudes,
        variances,
        image_to_walker_log_likelihood_fn,
        per_particle_args=per_particle_args_noise_variance,
        constant_args=1.0,
    )
    assert likelihood_matrix.shape == (n_particles, n_walkers)

    # test of _likelihood_sliced_wasserstein
    image_to_walker_log_likelihood_fn = likelihood_sliced_wasserstein
    per_particle_args_n_projections = 7

    n_particles = 5
    likelihood_matrix = compute_likelihood_matrix(
        ensemble_walkers,
        relion_dataset[:n_particles],
        amplitudes,
        variances,
        image_to_walker_log_likelihood_fn,
        constant_args=(per_particle_args_n_projections, 2),
        per_particle_args=(),
    )
    assert likelihood_matrix.shape == (n_particles, n_walkers)
"""
